#include "music.h"
#include <vector>
#include <math.h>
#include <algorithm>
#include <queue>

#include <iostream>
#include <fstream>

using namespace std;

// Generates array response vector.
Eigen::VectorXcd array_response_vector(std::vector<complex<double> >& array,
                                       double theta) {
    const std::complex< double> i(0, 2*M_PI*sin(theta));
    Eigen::VectorXcd v(array.size());
    for (int j = 0; j < array.size(); j++) {
        const std::complex< double> aa = i * array[j];
        v[j] = exp(aa);
    }

    return v/sqrt(array.size());
}

// MUSIC Algorithm.
vector<double> music(Eigen::MatrixXcd& CovMat, int L, int N,
                     vector<complex<double>>& array, vector<double>& Angles) {

    // Find eigenvectors.
    Eigen::SelfAdjointEigenSolver<Eigen::MatrixXcd> ces;
    ces.compute(CovMat);
    Eigen::MatrixXcd vecs= ces.eigenvectors();

    // Take necessary dimentions of the eigenvectos.
    Eigen::MatrixXcd qn(vecs.cols(),N-L) ;
    for (int i = 0; i < vecs.cols(); i++){
        for (int j = L; j < N ; j++) {
            qn.row(i)[j-L] = vecs.row(i)[vecs.cols()-j-1];
        }
    }

    int num_angles = Angles.size();
    vector<double> pspectrum(num_angles, 0) ;

    for (int i = 0; i < num_angles; i++) {
        Eigen::VectorXcd av = array_response_vector(array,Angles[i]);
        Eigen::VectorXcd ss = qn.conjugate().transpose() * av; 
        pspectrum[i] =  1/ss.norm();
    }

    // Get the min element of pspectrum.
    double min = pspectrum[0];
    for (int i = 0; i<pspectrum.size();i++) {
        if (pspectrum[i] < min) {
            min  = pspectrum[i];
        }
    }

    // Convert to log scale.
    vector<double> psindB(num_angles, 0) ;
    for (int i = 0; i<pspectrum.size();i++) {
        psindB[i] = log10(10*pspectrum[i]/min);
    }

    // Find the peaks of the pspectrum.
    std::vector<double> peaks;
    priority_queue<pair<double, double> > pq; 
    for (int i = 1; i<psindB.size()-1;i++) {
        if (psindB[i] >  psindB[i-1] && psindB[i] >  psindB[i+1] ) {
            pq.push({psindB[i], Angles[i]});
        }
    }

    // Get the corresponding Angle value of the largest peaks.
    for (int i = 0; i < L; i++) {
        peaks.push_back(pq.top().second);
        pq.pop();
    }

    // Sort the angles from smallest to largest.
    std::sort(peaks.begin(),peaks.end());
    return peaks;
}






